@@ -34,6 +34,7 @@
 #include <osmocom/core/logging.h>
 #include <osmocom/core/utils.h>
 #include <osmocom/core/rate_ctr.h>
+#include <osmocom/core/msgb.h>
 
 #include <osmocom/vty/logging.h>
 #include <osmocom/vty/telnet_interface.h>
@@ -41,10 +42,10 @@
 #include <osmocom/vty/misc.h>
 
 #include <osmocom/sgsn/debug.h>
-#include <osmocom/sgsn/gtphub.h>
+#include <osmocom/gtphub/gtphub.h>
 #include <osmocom/sgsn/vty.h>
 
-#include "../../bscconfig.h"
+#include "../../config.h"
 
 #if BUILD_IU
 #include <osmocom/sigtran/osmo_ss7.h>
@@ -353,8 +354,7 @@ int main(int argc, char **argv)
 	}
 
 	/* start telnet after reading config for vty_get_bind_addr() */
-	rc = telnet_init_dynif(osmo_gtphub_ctx, 0, vty_get_bind_addr(),
-			       OSMO_VTY_PORT_GTPHUB);
+	rc = telnet_init_default(osmo_gtphub_ctx, NULL, OSMO_VTY_PORT_GTPHUB);
 	if (rc < 0)
 		exit(1);
 
