@@ -33,7 +33,7 @@
 #include <osmocom/gsm/gsm_utils.h>
 
 #include <osmocom/sgsn/debug.h>
-#include <osmocom/sgsn/gprs_sgsn.h>
+#include <osmocom/sgsn/mmctx.h>
 #include <osmocom/sgsn/gprs_gmm.h>
 #include <osmocom/sgsn/gprs_llc.h>
 #include <osmocom/sgsn/crc24.h>
@@ -50,10 +50,21 @@ const struct value_string gprs_llc_llme_state_names[] = {
 	{ 0, NULL }
 };
 
+const struct value_string gprs_llc_lle_state_names[] = {
+	{ GPRS_LLES_UNASSIGNED,		"TLLI Unassigned" },
+	{ GPRS_LLES_ASSIGNED_ADM,	"TLLI Assigned" },
+	{ GPRS_LLES_LOCAL_EST,		"Local Establishment" },
+	{ GPRS_LLES_REMOTE_EST,		"Remote Establishment" },
+	{ GPRS_LLES_ABM,		"Asynchronous Balanced Mode" },
+	{ GPRS_LLES_LOCAL_REL,		"Local Release" },
+	{ GPRS_LLES_TIMER_REC,		"Timer Recovery" },
+	{ 0, NULL }
+};
+
 static struct gprs_llc_llme *llme_alloc(uint32_t tlli);
-static int gprs_llc_tx_xid(struct gprs_llc_lle *lle, struct msgb *msg,
+static int gprs_llc_tx_xid(const struct gprs_llc_lle *lle, struct msgb *msg,
 			   int command);
-static int gprs_llc_tx_dm(struct gprs_llc_lle *lle);
+static int gprs_llc_tx_dm(const struct gprs_llc_lle *lle);
 static int gprs_llc_tx_u(struct msgb *msg, uint8_t sapi,
 			 int command, enum gprs_llc_u_cmd u_cmd, int pf_bit);
 
@@ -212,7 +223,7 @@ static int gprs_llc_process_xid_ind(uint8_t *bytes_request,
 				    int bytes_request_len,
 				    uint8_t *bytes_response,
 				    int bytes_response_maxlen,
-				    struct gprs_llc_lle *lle)
+				    const struct gprs_llc_lle *lle)
 {
 	/* Note: This function computes the response that is sent back to the
 	 * MS when a mobile originated XID is received. The function is
@@ -288,7 +299,7 @@ static int gprs_llc_process_xid_ind(uint8_t *bytes_request,
 
 /* Dispatch XID indications and responses comming from the MS */
 static void rx_llc_xid(struct gprs_llc_lle *lle,
-		       struct gprs_llc_hdr_parsed *gph)
+		       const struct gprs_llc_hdr_parsed *gph)
 {
 	uint8_t response[1024];
 	int response_len;
@@ -681,7 +692,7 @@ int gprs_llc_tx_u(struct msgb *msg, uint8_t sapi, int command,
 }
 
 /* Send XID response to LLE */
-static int gprs_llc_tx_xid(struct gprs_llc_lle *lle, struct msgb *msg,
+static int gprs_llc_tx_xid(const struct gprs_llc_lle *lle, struct msgb *msg,
 			   int command)
 {
 	/* copy identifiers from LLE to ensure lower layers can route */
@@ -692,7 +703,7 @@ static int gprs_llc_tx_xid(struct gprs_llc_lle *lle, struct msgb *msg,
 	return gprs_llc_tx_u(msg, lle->sapi, command, GPRS_LLC_U_XID, 1);
 }
 
-static int gprs_llc_tx_dm(struct gprs_llc_lle *lle)
+static int gprs_llc_tx_dm(const struct gprs_llc_lle *lle)
 {
 	struct msgb *msg = msgb_alloc_headroom(4096, 1024, "LLC_DM");
 
@@ -705,7 +716,7 @@ static int gprs_llc_tx_dm(struct gprs_llc_lle *lle)
 }
 
 /* encrypt information field + FCS, if needed! */
-static int apply_gea(struct gprs_llc_lle *lle, uint16_t crypt_len, uint16_t nu,
+static int apply_gea(const struct gprs_llc_lle *lle, uint16_t crypt_len, uint16_t nu,
 		     uint32_t oc, uint8_t sapi, uint8_t *fcs, uint8_t *data)
 {
 	uint8_t cipher_out[GSM0464_CIPH_MAX_BLOCK];
@@ -965,9 +976,9 @@ int gprs_llc_rcvmsg(struct msgb *msg, struct tlv_parsed *tv)
 				       llhp.data);
 			if (rc < 0)
 				return rc;
-		llhp.fcs = *(llhp.data + llhp.data_len);
-		llhp.fcs |= *(llhp.data + llhp.data_len + 1) << 8;
-		llhp.fcs |= *(llhp.data + llhp.data_len + 2) << 16;
+			llhp.fcs = *(llhp.data + llhp.data_len);
+			llhp.fcs |= *(llhp.data + llhp.data_len + 1) << 8;
+			llhp.fcs |= *(llhp.data + llhp.data_len + 2) << 16;
 		} else {
 			LOGP(DLLC, LOGL_NOTICE, "encrypted frame for LLC that "
 				"has no KC/Algo! Dropping.\n");
@@ -1019,7 +1030,7 @@ int gprs_llc_rcvmsg(struct msgb *msg, struct tlv_parsed *tv)
 		case GPRS_SAPI_SNDCP9:
 		case GPRS_SAPI_SNDCP11:
 			/* send LL_DATA_IND/LL_UNITDATA_IND to SNDCP */
-			rc = sndcp_llunitdata_ind(msg, lle, llhp.data, llhp.data_len);
+			rc = sndcp_ll_unitdata_ind(msg, lle, llhp.data, llhp.data_len);
 			break;
 		case GPRS_SAPI_SMS:
 			/* FIXME */
@@ -1037,7 +1048,7 @@ int gprs_llc_rcvmsg(struct msgb *msg, struct tlv_parsed *tv)
 }
 
 /* Propagate crypto parameters MM -> LLME */
-void gprs_llme_copy_key(struct sgsn_mm_ctx *mm, struct gprs_llc_llme *llme)
+void gprs_llme_copy_key(const struct sgsn_mm_ctx *mm, struct gprs_llc_llme *llme)
 {
 	if (!mm)
 		return;
@@ -1101,7 +1112,7 @@ int gprs_llgmm_assign(struct gprs_llc_llme *llme,
 		llme->state = GPRS_LLMS_ASSIGNED;
 	} else if (old_tlli != TLLI_UNASSIGNED && new_tlli == TLLI_UNASSIGNED) {
 		/* TLLI Unassignment 8.3.3) */
-		llme->tlli = llme->old_tlli = 0;
+		llme->tlli = llme->old_tlli = TLLI_UNASSIGNED;
 		llme->state = GPRS_LLMS_UNASSIGNED;
 		for (i = 0; i < ARRAY_SIZE(llme->lle); i++) {
 			struct gprs_llc_lle *l = &llme->lle[i];
