@@ -22,7 +22,7 @@
  */
 
 #include "config.h"
-#include <osmocom/gtp/gtp.h>
+#include <gtp.h>
 
 #include <osmocom/core/rate_ctr.h>
 #include <osmocom/core/tdef.h>
@@ -137,17 +137,11 @@ ret_error:
 	return -1;
 }
 
-static int sgsn_ranap_iu_event_mmctx(struct ranap_ue_conn_ctx *ctx, enum ranap_iu_event_type type, void *data)
+int sgsn_ranap_iu_event(struct ranap_ue_conn_ctx *ctx, enum ranap_iu_event_type type, void *data)
 {
 	struct sgsn_mm_ctx *mm;
 	int rc = -1;
 
-	if (!ctx) {
-		LOGIUP(ctx, LOGL_ERROR, "NULL ctx given for IU event %s\n",
-		       ranap_iu_event_type_str(type));
-		return rc;
-	}
-
 	mm = sgsn_mm_ctx_by_ue_ctx(ctx);
 	if (!mm) {
 		LOGIUP(ctx, LOGL_NOTICE, "Cannot find mm ctx for IU event %s\n",
@@ -192,28 +186,13 @@ static int sgsn_ranap_iu_event_mmctx(struct ranap_ue_conn_ctx *ctx, enum ranap_i
 		rc = 0;
 		break;
 	default:
-		LOGMMCTXP(LOGL_NOTICE, mm, "Unknown event received: %d\n", type);
+		LOGMMCTXP(LOGL_NOTICE, mm, "Unknown event received: %i\n", type);
 		rc = -1;
 		break;
 	}
 	return rc;
 }
 
-
-int sgsn_ranap_iu_event(struct ranap_ue_conn_ctx *ctx, enum ranap_iu_event_type type, void *data)
-{
-	switch (type) {
-	case RANAP_IU_EVENT_RAB_ASSIGN:
-	case RANAP_IU_EVENT_IU_RELEASE:
-	case RANAP_IU_EVENT_LINK_INVALIDATED:
-	case RANAP_IU_EVENT_SECURITY_MODE_COMPLETE:
-		return sgsn_ranap_iu_event_mmctx(ctx, type, data);
-	default:
-		LOGP(DRANAP, LOGL_NOTICE, "Iu: Unknown event received: type: %d\n", type);
-		return -1;
-	}
-}
-
 void sgsn_ranap_iu_free(struct sgsn_mm_ctx *ctx)
 {
 	if (!ctx)
