@@ -24,6 +24,8 @@
 #include <arpa/inet.h>
 #include <time.h>
 #include <inttypes.h>
+#include <sys/stat.h>
+#include <sys/types.h>
 
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/utils.h>
@@ -35,11 +37,14 @@
 #include <osmocom/sgsn/debug.h>
 #include <osmocom/sgsn/sgsn.h>
 #include <osmocom/gprs/gprs_ns2.h>
-#include <osmocom/sgsn/gprs_gb.h>
+#include <osmocom/sgsn/gprs_ns.h>
 #include <osmocom/sgsn/gprs_gmm.h>
-#include <osmocom/sgsn/gprs_sgsn.h>
+#include <osmocom/sgsn/gprs_bssgp.h>
+#include <osmocom/sgsn/mmctx.h>
+#include <osmocom/sgsn/gtp_ggsn.h>
 #include <osmocom/sgsn/gtp_mme.h>
 #include <osmocom/sgsn/vty.h>
+#include <osmocom/sgsn/pdpctx.h>
 #include <osmocom/gsupclient/gsup_client.h>
 
 #include <osmocom/vty/tdef_vty.h>
@@ -48,21 +53,18 @@
 #include <osmocom/vty/misc.h>
 #include <osmocom/crypt/gprs_cipher.h>
 #include <osmocom/crypt/utran_cipher.h>
-#include <osmocom/abis/ipa.h>
 
 #include <osmocom/gprs/gprs_bssgp.h>
 
 #include <pdp.h>
 #include <gtp.h>
 
-#include "../../bscconfig.h"
+#include "../../config.h"
 
 #ifdef BUILD_IU
 #include <osmocom/ranap/iu_client.h>
 #endif
 
-extern void *tall_sgsn_ctx;
-
 static struct sgsn_config *g_cfg = NULL;
 
 const struct value_string sgsn_auth_pol_strs[] = {
@@ -97,7 +99,7 @@ const struct value_string sgsn_auth_pol_strs[] = {
 #define NONSPEC_X1001_SECS     5       /* wait for a RANAP Release Complete */
 
 
-static struct osmo_tdef sgsn_T_defs[] = {
+struct osmo_tdef sgsn_T_defs[] = {
 	{ .T=3312, .default_val=GSM0408_T3312_SECS, .desc="Periodic RA Update timer (s)" },
 	{ .T=3313, .default_val=GSM0408_T3313_SECS, .desc="Waiting for paging response timer (s)" },
 	{ .T=3314, .default_val=GSM0408_T3314_SECS, .desc="READY timer. Force to STANDBY on expiry timer (s)" },
@@ -136,6 +138,26 @@ DEFUN(cfg_sgsn_timer, cfg_sgsn_timer_cmd,
 	return osmo_tdef_vty_set_cmd(vty, g_cfg->T_defs, argv);
 }
 
+DEFUN(show_timer_gtp, show_timer_gtp_cmd,
+      "show timer gtp " OSMO_TDEF_VTY_ARG_T_OPTIONAL,
+      SHOW_STR "Show timers\n" "GTP (libgtp) timers\n"
+      OSMO_TDEF_VTY_DOC_T)
+{
+	const char *T_arg = argc > 0 ? argv[0] : NULL;
+	return osmo_tdef_vty_show_cmd(vty, g_cfg->T_defs_gtp, T_arg, NULL);
+}
+
+DEFUN(cfg_sgsn_timer_gtp, cfg_sgsn_timer_gtp_cmd,
+      "timer gtp " OSMO_TDEF_VTY_ARG_SET_OPTIONAL,
+      "Configure or show timers\n" "GTP (libgtp) timers\n"
+      OSMO_TDEF_VTY_DOC_SET)
+{
+	/* If any arguments are missing, redirect to 'show' */
+	if (argc < 2)
+		return show_timer(self, vty, argc, argv);
+	return osmo_tdef_vty_set_cmd(vty, g_cfg->T_defs_gtp, argv);
+}
+
 char *gprs_pdpaddr2str(uint8_t *pdpa, uint8_t len, bool return_ipv6)
 {
 	static char str[INET6_ADDRSTRLEN + 10];
@@ -230,7 +252,7 @@ static int config_write_sgsn(struct vty *vty)
 	vty_out(vty, " gtp local-ip %s%s",
 		inet_ntoa(g_cfg->gtp_listenaddr.sin_addr), VTY_NEWLINE);
 
-	llist_for_each_entry(gctx, &sgsn_ggsn_ctxts, list) {
+	llist_for_each_entry(gctx, &sgsn->ggsn_list, list) {
 		if (gctx->id == UINT32_MAX)
 			continue;
 
@@ -261,6 +283,17 @@ static int config_write_sgsn(struct vty *vty)
 
 		vty_out(vty, "%s", VTY_NEWLINE);
 	}
+	if (g_cfg->uea_encryption_mask != 0) {
+		vty_out(vty, " encryption uea");
+
+		for (i = 0; i < _OSMO_UTRAN_UEA_NUM; i++)
+			if (g_cfg->uea_encryption_mask >> i & 1)
+				vty_out(vty, " %u", i);
+
+		vty_out(vty, "%s", VTY_NEWLINE);
+	}
+	if (g_cfg->crypt_cipher_plugin_path)
+		vty_out(vty, " encryption cipher-plugin-path %s%s", g_cfg->crypt_cipher_plugin_path, VTY_NEWLINE);
 	if (g_cfg->sgsn_ipa_name)
 		vty_out(vty, " gsup ipa-name %s%s", g_cfg->sgsn_ipa_name, VTY_NEWLINE);
 	if (g_cfg->gsup_server_addr.sin_addr.s_addr)
@@ -289,9 +322,9 @@ static int config_write_sgsn(struct vty *vty)
 	llist_for_each_entry(acl, &g_cfg->imsi_acl, list)
 		vty_out(vty, " imsi-acl add %s%s", acl->imsi, VTY_NEWLINE);
 
-	if (llist_empty(&sgsn_apn_ctxts))
+	if (llist_empty(&sgsn->apn_list))
 		vty_out(vty, " ! apn * ggsn 0%s", VTY_NEWLINE);
-	llist_for_each_entry(actx, &sgsn_apn_ctxts, list) {
+	llist_for_each_entry(actx, &sgsn->apn_list, list) {
 		if (strlen(actx->imsi_prefix) > 0)
 			vty_out(vty, " apn %s imsi-prefix %s ggsn %u%s",
 				actx->name, actx->imsi_prefix, actx->ggsn->id,
@@ -312,6 +345,7 @@ static int config_write_sgsn(struct vty *vty)
 	vty_out(vty, " cdr interval %d%s", g_cfg->cdr.interval, VTY_NEWLINE);
 
 	osmo_tdef_vty_write(vty, g_cfg->T_defs, " timer ");
+	osmo_tdef_vty_write(vty, g_cfg->T_defs_gtp, " timer gtp ");
 
 	if (g_cfg->pcomp_rfc1144.active) {
 		vty_out(vty, " compression rfc1144 active slots %d%s",
@@ -371,6 +405,11 @@ DEFUN(cfg_sgsn_state_dir, cfg_sgsn_state_dir_cmd,
 	"Set the directory for the GTP State file\n"
 	"Local Directory\n")
 {
+	if (mkdir(argv[0], 0755) == -1 && errno != EEXIST) {
+		vty_out(vty, "%% Failed to create state-dir: %s%s", argv[0], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
 	osmo_talloc_replace_string(sgsn, &sgsn->cfg.gtp_statedir, argv[0]);
 
 	return CMD_SUCCESS;
@@ -399,7 +438,7 @@ DEFUN(cfg_ggsn_remote_ip, cfg_ggsn_remote_ip_cmd,
 	"IPv4 Address\n")
 {
 	uint32_t id = atoi(argv[0]);
-	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(id);
+	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(sgsn, id);
 
 	inet_aton(argv[1], &ggc->remote_addr);
 
@@ -424,7 +463,7 @@ DEFUN(cfg_ggsn_gtp_version, cfg_ggsn_gtp_version_cmd,
 	"Version 0\n" "Version 1\n")
 {
 	uint32_t id = atoi(argv[0]);
-	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(id);
+	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(sgsn, id);
 
 	if (atoi(argv[1]))
 		ggc->gtp_version = 1;
@@ -442,7 +481,7 @@ DEFUN(cfg_ggsn_echo_interval, cfg_ggsn_echo_interval_cmd,
 	"Interval between echo requests in seconds.\n")
 {
 	uint32_t id = atoi(argv[0]);
-	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(id);
+	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(sgsn, id);
 
 	ggc->echo_interval = atoi(argv[1]);
 
@@ -461,7 +500,7 @@ DEFUN(cfg_ggsn_no_echo_interval, cfg_ggsn_no_echo_interval_cmd,
 	NO_STR "Send an echo request to this static GGSN every interval.\n")
 {
 	uint32_t id = atoi(argv[0]);
-	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(id);
+	struct sgsn_ggsn_ctx *ggc = sgsn_ggsn_ctx_find_alloc(sgsn, id);
 
 	ggc->echo_interval = 0;
 	sgsn_ggsn_ctx_check_echo_timer(ggc);
@@ -502,7 +541,7 @@ static int add_apn_ggsn_mapping(struct vty *vty, const char *apn_str,
 	struct apn_ctx *actx;
 	struct sgsn_ggsn_ctx *ggsn;
 
-	ggsn = sgsn_ggsn_ctx_by_id(ggsn_id);
+	ggsn = sgsn_ggsn_ctx_by_id(sgsn, ggsn_id);
 	if (ggsn == NULL) {
 		vty_out(vty, "%% a GGSN with id %d has not been defined%s",
 			ggsn_id, VTY_NEWLINE);
@@ -611,7 +650,7 @@ static void vty_dump_mmctx(struct vty *vty, const char *pfx,
 		pfx, mm->msisdn, id, mm->hlr, VTY_NEWLINE);
 	vty_out(vty, "%s  GMM State: %s, Routeing Area: %s, Cell ID: %u%s",
 		pfx, osmo_fsm_inst_state_name(mm->gmm_fsm),
-		osmo_rai_name(&mm->ra), mm->gb.cell_id, VTY_NEWLINE);
+		osmo_rai_name2(&mm->ra), mm->gb.cell_id, VTY_NEWLINE);
 	vty_out(vty, "%s  MM State: %s, RAN Type: %s%s", pfx, mm_state_name,
 		get_value_string(sgsn_ran_type_names, mm->ran_type), VTY_NEWLINE);
 
@@ -629,11 +668,10 @@ DEFUN(show_sgsn, show_sgsn_cmd, "show sgsn",
       SHOW_STR "Display information about the SGSN")
 {
 	if (sgsn->gsup_client) {
-		struct ipa_client_conn *link = sgsn->gsup_client->link;
-		vty_out(vty,
-			"  Remote authorization: %sconnected to %s:%d via GSUP%s",
-			sgsn->gsup_client->is_connected ? "" : "not ",
-			link->addr, link->port,
+		vty_out(vty, "  Remote authorization: %sconnected to %s:%d via GSUP%s",
+			osmo_gsup_client_is_connected(sgsn->gsup_client) ? "" : "not ",
+			osmo_gsup_client_get_rem_addr(sgsn->gsup_client),
+			osmo_gsup_client_get_rem_port(sgsn->gsup_client),
 			VTY_NEWLINE);
 	}
 	if (sgsn->gsn)
@@ -689,7 +727,7 @@ DEFUN(swow_mmctx_all, show_mmctx_all_cmd,
 	SHOW_STR MMCTX_STR "All MM Contexts\n" INCLUDE_PDP_STR)
 {
 	struct sgsn_mm_ctx *mm;
-	llist_for_each_entry(mm, &sgsn_mm_ctxts, list)
+	llist_for_each_entry(mm, &sgsn->mm_list, list)
 		vty_dump_mmctx(vty, "", mm, (argc > 0) ? 1 : 0);
 
 	return CMD_SUCCESS;
@@ -701,7 +739,7 @@ DEFUN(show_pdpctx_all, show_pdpctx_all_cmd,
 {
 	struct sgsn_pdp_ctx *pdp;
 
-	llist_for_each_entry(pdp, &sgsn_pdp_ctxts, g_list)
+	llist_for_each_entry(pdp, &sgsn->pdp_list, g_list)
 		vty_dump_pdp(vty, "", pdp);
 
 	return CMD_SUCCESS;
@@ -817,6 +855,27 @@ DEFUN(cfg_encrypt2, cfg_encrypt2_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_encrypt_cipher_plugin_path, cfg_encrypt_cipher_plugin_path_cmd,
+	"encryption cipher-plugin-path PATH",
+	ENCRYPTION_STR
+	"Path to gprs encryption cipher plugin directory\n"
+	"Plugin path\n")
+{
+	osmo_talloc_replace_string(sgsn, &sgsn->cfg.crypt_cipher_plugin_path, argv[0]);
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_no_encrypt_cipher_plugin_path, cfg_no_encrypt_cipher_plugin_path_cmd,
+	"no encryption cipher-plugin-path PATH",
+	NO_STR ENCRYPTION_STR
+	"Path to gprs encryption cipher plugin directory\n"
+	"Plugin path\n")
+{
+	TALLOC_FREE(sgsn->cfg.crypt_cipher_plugin_path);
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_authentication, cfg_authentication_cmd,
       "authentication (optional|required)",
       "Whether to enforce MS authentication in GERAN (only with auth-policy remote)\n"
@@ -936,8 +995,17 @@ static void subscr_dump_full_vty(struct vty *vty, struct gprs_subscr *gsub, int
 	}
 
 	llist_for_each_entry(pdp, &gsub->sgsn_data->pdp_list, list) {
-		vty_out(vty, "    PDP info: Id: %d, Type: 0x%04x, APN: '%s'",
-			pdp->context_id, pdp->pdp_type, pdp->apn_str);
+		char ip_str[INET6_ADDRSTRLEN] = { 0 };
+
+		vty_out(vty, "    PDP info: Id: %d, Addr(Org: 0x%02x Type: 0x%02x",
+			pdp->context_id, pdp->pdp_type_org, pdp->pdp_type_nr);
+
+		if (pdp->pdp_address[0].u.sa.sa_family != AF_UNSPEC)
+			vty_out(vty, " Addr[0]: %s", osmo_sockaddr_ntop(&pdp->pdp_address[0].u.sa, ip_str));
+		if (pdp->pdp_address[0].u.sa.sa_family != AF_UNSPEC)
+			vty_out(vty, " Addr[1]: %s", osmo_sockaddr_ntop(&pdp->pdp_address[1].u.sa, ip_str));
+
+		vty_out(vty, ") APN: '%s'", pdp->apn_str);
 
 		if (pdp->qos_subscribed_len)
 			vty_out(vty, " QoS: %s", osmo_hexdump(pdp->qos_subscribed, pdp->qos_subscribed_len));
@@ -984,7 +1052,7 @@ DEFUN_HIDDEN(reset_sgsn_state,
 	struct gprs_subscr *subscr, *tmp_subscr;
 	struct sgsn_mm_ctx *mm, *tmp_mm;
 
-	llist_for_each_entry_safe(mm, tmp_mm, &sgsn_mm_ctxts, list)
+	llist_for_each_entry_safe(mm, tmp_mm, &sgsn->mm_list, list)
 	{
 		gsm0408_gprs_access_cancelled(mm, SGSN_ERROR_CAUSE_NONE);
 	}
@@ -1263,7 +1331,7 @@ DEFUN(page_subscr, page_subscr_info_cmd,
 		return CMD_WARNING;
 	}
 
-	gprs_gb_page_ps_ra(mm);
+	sgsn_bssgp_page_ps_ra(mm);
 	return CMD_SUCCESS;
 }
 
@@ -1716,9 +1784,6 @@ int sgsn_vty_init(struct sgsn_config *cfg)
 {
 	g_cfg = cfg;
 
-	g_cfg->T_defs = sgsn_T_defs;
-	osmo_tdefs_reset(g_cfg->T_defs);
-
 	install_element_ve(&show_sgsn_cmd);
 	//install_element_ve(&show_mmctx_tlli_cmd);
 	install_element_ve(&show_mmctx_imsi_cmd);
@@ -1726,6 +1791,7 @@ int sgsn_vty_init(struct sgsn_config *cfg)
 	install_element_ve(&show_pdpctx_all_cmd);
 	install_element_ve(&show_subscr_cache_cmd);
 	install_element_ve(&show_timer_cmd);
+	install_element_ve(&show_timer_gtp_cmd);
 
 	install_element(ENABLE_NODE, &update_subscr_insert_auth_triplet_cmd);
 	install_element(ENABLE_NODE, &update_subscr_create_cmd);
@@ -1753,6 +1819,8 @@ int sgsn_vty_init(struct sgsn_config *cfg)
 	install_element(SGSN_NODE, &cfg_encrypt2_cmd);
 	install_element(SGSN_NODE, &cfg_encrypt_cmd);
 	install_element(SGSN_NODE, &cfg_encryption_uea_cmd);
+	install_element(SGSN_NODE, &cfg_encrypt_cipher_plugin_path_cmd);
+	install_element(SGSN_NODE, &cfg_no_encrypt_cipher_plugin_path_cmd);
 
 	install_element(SGSN_NODE, &cfg_gsup_ipa_name_cmd);
 	install_element(SGSN_NODE, &cfg_gsup_remote_ip_cmd);
@@ -1773,6 +1841,7 @@ int sgsn_vty_init(struct sgsn_config *cfg)
 	install_element(SGSN_NODE, &cfg_grx_ggsn_cmd);
 
 	install_element(SGSN_NODE, &cfg_sgsn_timer_cmd);
+	install_element(SGSN_NODE, &cfg_sgsn_timer_gtp_cmd);
 
 	install_element(SGSN_NODE, &cfg_no_comp_rfc1144_cmd);
 	install_element(SGSN_NODE, &cfg_comp_rfc1144_cmd);
@@ -1804,9 +1873,6 @@ int sgsn_parse_config(const char *config_file)
 	/* make sure sgsn_vty_init() was called before this */
 	OSMO_ASSERT(g_cfg);
 
-	g_cfg->gea_encryption_mask = 0x1; /* support GEA0 by default unless specific encryption config exists */
-	g_cfg->uea_encryption_mask = (1 << OSMO_UTRAN_UEA0); /* support UEA0 by default unless specific encryption config exists */
-
 	rc = vty_read_config_file(config_file, NULL);
 	if (rc < 0) {
 		fprintf(stderr, "Failed to parse the config file: '%s'\n", config_file);
